#!/usr/bin/env python
# -*- coding: utf-8 -*-
# PYTHON_ARGCOMPLETE_OK
#
#    Timetra is a time tracking application and library.
#    Copyright © 2010-2014  Andrey Mikhaylenko
#
#    This file is part of Timetra.
#
#    Timetra is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Lesser General Public License as published
#    by the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    Timetra is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public License
#    along with Timetra.  If not, see <http://gnu.org/licenses/>.
#
"""
======================
Command-Line Interface
======================

:author: Andrey Mikhaylenko

"""
import datetime
import re
import textwrap

from argh import (aliases, arg, confirm, CommandError, expects_obj,
                  named, wrap_errors)
from argh.io import safe_input
from confu import Configurable
import prettytable
import yaml

from .term import success, warning, failure, t
from . import models, utils, formatdelta
from .storage import Storage, StorageError


TIMETRA_TAG = 'timetra'
TIMETRA_TAG_LOG = 'timetra-log'
TIMETRA_TAG_LOAD = 'timetra-load'


class NotFoundError(Exception):
    pass


def parse_activity(storage, loose_name):
    try:
        return storage.parse_activity(loose_name)
    except storage.ActivityMatchingError as e:
        raise CommandError(failure(e))


def _get_last_fact(storage, activity_mask=None, days=2):
    if not activity_mask:
        return storage.get_latest_fact()

    activity = storage.resolve_activity(activity_mask)['activity']
    until = datetime.datetime.now()
    since = until - datetime.timedelta(days=days)
    facts = storage.find(since, until, activity)

    fact = facts[-1] if facts else None

    if fact:
        assert fact.activity == activity, fact.activity + '|'+ activity
    else:
        obj = 'a "{0}" fact'.format(activity) if activity else 'a fact'
        raise NotFoundError('Could not find {0} within last '
                            '{1} days'.format(obj, days))
    return fact


def punch_in(storage, activity, continued=False, interactive=False):
    """Starts tracking given activity. Stops tracking on C-c.

    :param continued:

        The start time is taken from the last logged fact's end time. If that
        fact is not marked as finished, it is ended now. If it describes the
        same activity and is not finished, it is continued; if it is already
        finished, user is prompted for action.

    :param interactive:

        In this mode the application prompts for user input, adds it to the
        fact description (with timestamp) and displays the prompt again. The
        first empty comment stops current activity and terminates the app.

        Useful for logging work obstacles, decisions, ideas, etc.

    """
    # TODO:
    # * smart "-c":
    #   * "--upto DURATION" modifier (avoids overlapping)
    activity, category = parse_activity(activity)
    start = None
    fact = None
    if continued:
        prev = storage.get_latest()
        if prev:
            if prev.activity == activity and prev.category == category:
                do_cont = True
                #comment = None
                if prev.end_time:
                    delta = datetime.datetime.now() - prev.end_time
                    question = (u'Merge with previous entry filling {0} of '
                                 'inactivity'.format(utils.format_delta(delta)))
                    if not confirm(question, default=True):
                        do_cont = False
                    #comment = question

                if do_cont:
                    fact = prev
                    storage.update(fact, {'until': None})

            # if the last activity has not ended yet, it's ok: the `start`
            # variable will be `None`
            start = prev.end_time
            if start:
                yield u'Logging activity as started at {0}'.format(start)

    if not fact:
        fact = models.Fact(activity=activity, category=category,
                           tags=[TIMETRA_TAG], since=start)
        storage.add(fact)
        for line in show_last_fact():
            yield line

    if not interactive:
        return

    yield u'Type a comment and hit Enter. Empty comment ends activity.'
    try:
        while True:
            comment = raw_input(u'-> ').strip()
            if not comment:
                break
            # FIXME this is currently broken
            fact = storage.get_current_fact()
            assert fact, 'all logged activities are already closed'
            storage.update(fact, extra_description=comment)
    except KeyboardInterrupt:
        pass

    # FIXME this is currently broken
    storage.stop_tracking()
    for line in show_last_fact():
        yield line


def punch_out(storage, description=None, tags=None, with_person=None):
    "Stops an ongoing activity tracking (i.e. closes the latest activity)."
    kwargs = {}

    if description:
        kwargs.update(extra_description=description)

    # tags
    extra_tags = []
    if tags:
        extra_tags.extend(tags.split(','))
    if with_person:
        extra_tags.extend(['with-{0}'.format(x) for x in with_person.split(',')])
    if extra_tags:
        kwargs.update(extra_tags=extra_tags)

    fact = storage.get_current_fact()

    if not fact:
        raise CommandError(failure(u'No activity is running.'))

    if kwargs:
        storage.update_fact(fact, **kwargs)

    storage.stop_tracking()

    for line in show_last_fact():
        yield line

def log_activity(args):
    "Logs a past activity (since last logged until now)"

    # TODO: split all changes into steps and apply each using a separate,
    # well-tested function

    since = args.since
    until = args.until
    duration = args.duration

    if not args.activity and not args.amend:
        raise CommandError(failure('activity must be specified '
                                   'unless --amend is set'))

    if args.between:
        assert not (since or until or duration), (
            '--since, --until and --duration must not be used with --between')
        since, until = args.between.split('-')

    if args.date:
        rel_date = datetime.datetime.strptime(args.date, '%Y-%m-%d')
        since = utils.parse_time_to_datetime(since, relative_to=rel_date,
                                             ensure_past_time=False)
        until = utils.parse_time_to_datetime(until, relative_to=rel_date,
                                             ensure_past_time=False)
    else:
        since = utils.parse_time_to_datetime(since)
        until = utils.parse_time_to_datetime(until)

    delta = utils.parse_delta(duration)

    tags = [TIMETRA_TAG_LOG]
    if args.tags:
        tags = list(set(tags + args.tags.split(',')))
    if args.with_person:
        tags.extend(['with-{0}'.format(x) for x in args.with_person.split(',')])

    if args.pick and not args.amend:
        raise CommandError(failure('--pick only makes sense with --amend'))

    prev = _get_last_fact(args.pick)

    if args.amend:
        if not prev:
            raise CommandError('Cannot amend: no fact found')

        # FIXME this disables --duration
        since = since or prev.start_time
        until = until or prev.end_time

    start, end = storage.get_start_end(since, until, delta)

    if end < start:
        raise CommandError('--since must be earlier than --until')

    if datetime.datetime.now() < end:
        raise CommandError('--until must not be in the future')

    # check if we aren't going to overwrite any previous facts
    try:
        for line in check_overlap(start, end,
                                  activity=(args.pick or args.activity),
                                  amend_fact=prev if args.amend else None):
            yield line
    except OverlapError as e:
        raise CommandError(failure(e))

    if args.activity:
        activity, category = parse_activity(args.activity)
    else:
        activity = category = None

    description = None
    if args.description:
        description = args.description
    elif args.no_input:
        pass
    elif args.amend and prev.description:
        # updating a fact that already has a description
        pass
    else:
        # collect multi-line description from interactive user input
        lines = []
        num = 0
        try:
            while True:
                line = safe_input('        > ' if num else warning('Describe> '))
                if line:
                    lines.append(line)
                else:
                    yield ''
                    break
                num += 1
        except (KeyboardInterrupt, EOFError):
            raise CommandError(failure('Operation cancelled.'))
        else:
            description = '\n'.join(lines) if lines else None

    if args.amend:
        #template = u'Updated {fact.activity}@{fact.category} ({delta_minutes} min)'
        assert prev
        fact = prev

        kwargs = dict(
            start_time=start,
            end_time=end,
            dry_run=args.dry_run,
        )
        if activity:
            kwargs.update(activity=activity, category=category)
        if description is not None:
            kwargs.update(description=description)
        if args.tags is not None or args.with_person is not None:
            kwargs.update(tags=tags)

        changed = []
        for key, value in kwargs.iteritems():
            if hasattr(fact, key) and getattr(fact, key) != kwargs[key]:
                changed.append(key)
                old_value = getattr(fact, key)
                if hasattr(old_value, '__iter__'):
                    # convert DBus strings to proper pythonic ones (for tags)
                    old_value = [str(x) for x in old_value]
                note = u''
                if isinstance(old_value, datetime.datetime) and value:
                    if old_value < value:
                        note = u'(+{0})'.format(value - old_value)
                    else:
                        note = u'(-{0})'.format(old_value - value)
                yield u'* {0}: {1} →  {2} {3}'.format(
                    key,
                    failure(unicode(old_value)),
                    success(unicode(value)),
                    note)

        if not changed:
            yield failure(u'Nothing changed.')
            return

        storage.update_fact(fact, **kwargs)
    else:
        #template = u'Logged {fact.activity}@{fact.category} ({delta_minutes} min)'
        try:
            fact = storage.add_fact(
                args.activity,    # NOTE: not using parsed activity + category because hamster wants the foo@bar thing
                start_time=start,
                end_time=end,
                description=description,
                tags=tags,
                dry_run=args.dry_run)
        except (storage.ActivityMatchingError, storage.CannotCreateFact) as e:
            raise CommandError(failure(e))

    # report
    #delta = fact.end_time - start  # почему-то сам факт "не знает" времени начала
    #delta_minutes = delta.seconds / 60
    #yield success(template.format(fact=fact, delta_minutes=delta_minutes))

    for output in show_last_fact(args.activity or args.pick):
        yield output

    if args.dry_run:
        yield warning(u'(Dry run, nothing changed.)')


class OverlapError(RuntimeError):
    pass


class TooManyOverlappingFacts(OverlapError):
    pass


class FactOverlapsReplacement(OverlapError):
    pass


def check_overlap(start, end, activity='NEW ACTIVITY', amend_fact=None):
    """ Interactive check for overlapping facts.  To be used from other
    commands as generator.
    """
    def overlaps(fact, start_time, end_time):
        if not fact.end_time:
            # previous activity is still open
            return True
        if start_time >= fact.end_time or end_time <= fact.start_time:
            return False
        return True

    # check if we aren't going to overwrite any previous facts
    # FIXME not today but start.date() .. end.date()
    todays_facts = storage.get_facts_for_day(
        date =   (start - datetime.timedelta(days=1)).date(),
        end_date = (end + datetime.timedelta(days=1)).date())

    overlap = [f for f in todays_facts if overlaps(f, start, end)]

    if amend_fact:
        # do not count last fact as overlapping if we are about to change it.
        # using unicode(fact) because Hamster's Fact objects cannot be compared
        # directly for some reason.
        overlap = [f for f in overlap if not unicode(f) == unicode(amend_fact)]

    if not overlap:
        return

    if 1 < len(overlap):
        raise TooManyOverlappingFacts('FAIL: too many overlapping facts')

    prev_fact = overlap[-1]

    if start <= prev_fact.start_time and prev_fact.end_time <= end:
        # new fact devours an older one; this cannot be handled "properly"
        # FIXME: should count deltas <1min as equality
        raise FactOverlapsReplacement('FAIL: new fact would replace an older one')

    # FIXME: probably time should be rounded to seconds or even minutes
    #        for safer comparisons (i.e. 15:30:15 == 15:30:20)

    #--- begin vision   (pure visualization; backend will make decisions
    #                    on its own, hopefully in the same vein)

    outcome = []
    old = prev_fact.activity
    new = activity

    if prev_fact.start_time < start:
        outcome.append((warning(old), start - prev_fact.start_time))
    outcome.append((success(new), end - start))
    if end < prev_fact.end_time:
        outcome.append((warning(old), prev_fact.end_time - end))

    vision = '  '.join(u'[{0} +{1}]'.format(x[0], utils.format_delta(x[1])) for x in outcome)

    yield u'Before:  [{0} +{1}]'.format(failure(prev_fact.activity),
                                        utils.format_delta(prev_fact.delta))
    yield u' After:  {0}'.format(vision)

    #
    #--- end vision

    if not confirm(u'OK', default=False):
        raise CommandError('Operation cancelled.')


@aliases('ps')
def add_post_scriptum(*text):
    "Adds given text to the last logged (or current) fact."
    assert text
    fact = storage.get_latest_fact()
    assert fact
    text = ' '.join(text)
    storage.update_fact(fact, extra_description=text)

    for output in show_last_fact():
        yield output


@named('find')
@arg('query', help='"," = OR, " " = AND')
# NOTE: alas, Hamster does not support precise search by fields
#@arg('-c', '--category')
#@arg('-a', '--activity')
#@arg('-d', '--description')
#@arg('-t', '--tags')
@arg('-d', '--days', help='number of days to examine')
@arg('--summary', help='display only summary')
@arg('--show-date-if-crosses-days', help='display full end date if event '
                                         'spans multiple days')
def find_facts(storage, query, days=1, summary=False, show_date_if_crosses_days=False, compact=False):
    "Queries the fact database."
    until = datetime.datetime.now()
    since = until - datetime.timedelta(days=days)
    yield '# query "{0}"'.format(query)
    yield '# since {0}'.format(since)
    yield '# until {0}'.format(until)
    yield ''
    facts = storage.get_facts_for_day(since, end_date=until,
                                      search_terms=query)
    first_notion = None
    total_spent = datetime.timedelta()
    total_found = 0
    min_duration = None
    min_duration_event = None
    max_duration = None
    max_duration_event = None
    seen_workdays = {}
    last_date = None

    def make_table():
        tbl = prettytable.PrettyTable()
        if compact:
            tbl.field_names = ['time', 'activity', 'summary']
        else:
            tbl.field_names = ['time', 'activity', 'delta', 'summary', 'tags']
        tbl.align = 'l'
        return tbl

    table = None

    for fact in facts:
        if not summary:
            start = fact.start_time
            since_repr = start.strftime('%H:%M')
            if not last_date or last_date != start.date():
                if table:
                    yield table
                yield ''
                yield t.blue(start.strftime('%d %b %Y'))
                yield ''
                table = make_table()
            last_date = start.date()

            until_repr = fact.end_time.strftime('%H:%M')
            if (fact.start_time.date() != fact.end_time.date()
                and show_date_if_crosses_days):
                until_repr = fact.end_time.strftime('%Y-%m-%d %H:%M')

            tags = (unicode(x) for x in fact.tags)
            tags = [x for x in tags if not x in (TIMETRA_TAG, TIMETRA_TAG_LOG)]
            if compact:
                activity_repr = u'{0.activity}'.format(fact)
                description_width = 40
            else:
                activity_repr = u'{0.category}/{0.activity}'.format(fact)
                description_width = 70
            time_repr = '{}—{}'.format(since_repr, until_repr)
            tags_repr = u'#{0}'.format(' #'.join(tags)) if tags else '—'
            summary_repr = textwrap.fill(fact.description or '—',
                                         width=description_width)
            delta_repr = '+{0}'.format(utils.format_delta(fact.delta))

            if compact:
                summary_repr += u' {}'.format(tags_repr)
                time_repr += u' {}'.format(delta_repr)
                row = [time_repr, activity_repr, summary_repr]
            else:
                row = [time_repr, activity_repr, delta_repr, summary_repr, tags_repr]

            table.add_row(row)

        if not first_notion:
            first_notion = fact.start_time
        total_spent += fact.delta
        total_found += 1
        if min_duration is None or (datetime.timedelta(minutes=1) < fact.delta and fact.delta < min_duration):
            min_duration = fact.delta
            min_duration_event = fact
        if max_duration is None or max_duration < fact.delta:
            max_duration = fact.delta
            max_duration_event = fact
        seen_workdays[fact.start_time.date()] = 1

    if not summary:
        yield table

    if not total_found:
        yield failure(u'No facts found.')
        return

    total_workdays = len(seen_workdays)
    yield ''
    yield u'# Summary'
    yield u''
    yield u'* {0} facts'.format(warning(total_found))
    yield u'* {0} was the first notion ({1} ago)'.format(
        warning(first_notion), datetime.datetime.now() - first_notion)
    yield u'* {0} spent in total'.format(warning(total_spent))
    total_minutes = total_spent.total_seconds() / 60
    total_hours = total_minutes / 60
    yield u'* duration:'
    yield u'  {0:.0f} minutes ({1:.1f} hours)  per event'.format(
        total_minutes / (total_found or 1), total_hours / (total_found or 1))
    yield u'  {0:.0f} minutes ({1:.1f} hours)  per day'.format(
        total_minutes / days, total_hours / days)
    # "workdays" here are dates when given activity was started at least once.
    yield u'  {0:.0f} minutes ({1:.1f} hours)  per workday'.format(
        total_minutes / (total_workdays or 1),
        total_hours / (total_workdays or 1))
    if min_duration:
        yield u'  {0}  min event duration ({1.start_time})'.format(
            utils.format_delta(min_duration), min_duration_event)
    if max_duration:
        yield u'  {0}  max event duration ({1.start_time})'.format(
            utils.format_delta(max_duration), max_duration_event)


@named('last')
@arg('activity_mask', nargs='?', help='activity name (short form)')
@arg('--days', help='if `activity` is given, search this deep')
@arg('-v', '--verbose', default=False)
def show_last_fact(activity_mask=None, days=365, verbose=False):
    "Displays short note about current or latest activity, if any."

    fact = _get_last_fact(activity_mask, days=days)

    if not fact:
        yield u'--'
        return

    if fact.end_time:
        gap = datetime.datetime.now() - fact.end_time
        if gap.total_seconds() < 60:
            chart_right = u']  just finished'
        elif gap.total_seconds() < 60*24:
            chart_right = u']  ... +{0}'.format(utils.format_delta(gap))
        else:
            chart_right = u']  ... +{0}'.format(gap)
    else:
        chart_right = u'...>'
    yield u'{start}  [ {name}  +{duration} {right}'.format(
        name=success(fact.activity),
        start=fact.start_time.strftime('%H:%M'),
        duration=utils.format_delta(fact.delta),
        right=chart_right
    )
    if fact.description:
        yield u''
        yield u'\n'.join(u'       {0}'.format(x) for x in fact.description.split('\n'))

    if not verbose:
        return

    yield u''
    padding = max(len(k) for k in fact.__dict__)
    field_template = u'{key:>{padding}}: {value}'
    for k in sorted(fact.__dict__):
        value = getattr(fact, k)
        if k == 'tags':
            value = ', '.join(unicode(tag) for tag in value)
        yield field_template.format(key=k, value=value, padding=padding)


@named('update')
@arg('-n', '--number', help='number of the fact: latest is 1, previous 2, etc')
def update_fact(number=1, activity=None):
    latest_facts = storage.get_facts_for_day()
    fact = latest_facts[-number]
    kwargs = {}
    if activity:
        yield u'Updating fact {0}'.format(fact)
        activity, category = parse_activity(activity)
        kwargs['activity'] = activity
        kwargs['category'] = category
        storage.update_fact(fact, **kwargs)
    else:
        yield failure(u'No arguments given.')


@aliases('load')
def load_from_file(path, dry_run=False):
    """
    Assumes the following format::

        2013-01-27:
            - "0100-0130 activity description #with #tags"

    """
    with open(path) as f:
        dates = yaml.load(f)

    def _parse_record(raw, date):
        date = datetime.datetime.combine(date, datetime.time(23,59))
        pattern = re.compile(r'(?P<since>\d{4})\-(?P<until>\d{4})\s'
                             r'(?P<activity>[0-9a-z\-_@]+)'
                             r'(?P<description>\s+.+)?')
        match = pattern.search(raw)
        if not match:
            raise ValueError(u'could not parse "{0}"'.format(raw))
        gd = match.groupdict()
        since = utils.parse_time_to_datetime(gd['since'], relative_to=date,
                                             ensure_past_time=False)
        until = utils.parse_time_to_datetime(gd['until'], relative_to=date,
                                             ensure_past_time=False)
        if until < since:
            # started on one day, ended on another
            until += datetime.timedelta(days=1)
        activity, category = parse_activity(gd['activity'])
        description = gd['description'] or ''

        # partially mimic Hamster's tag-in-description serialization convention
        if ' #' in description:
            description, tags = description.split(' #', 1)
            tags = [t.strip() for t in tags.split('#') if t.strip()]
        else:
            tags = []
        tags.append(TIMETRA_TAG_LOAD)

        return {
            'since': since,
            'until': until,
            'activity': '@'.join([activity, category]),
            'description': description.strip(),
            'tags': tags
        }

    for date in sorted(dates):
        yield date
        for record in dates[date]:
            yield 'IN: {0}'.format(record)
            fact_data = _parse_record(record, date)
            try:
                for line in check_overlap(fact_data['since'],
                                          fact_data['until'],
                                          activity=fact_data['activity']):
                    yield line
            except FactOverlapsReplacement:
                yield '...already imported, skipping'
            else:
                fact = storage.add_fact(
                    fact_data['activity'],
                    start_time=fact_data['since'],
                    end_time=fact_data['until'],
                    description=fact_data['description'],
                    tags=fact_data['tags'],
                    dry_run=dry_run)
                yield 'SAVED: {0}'.format(fact)
        yield '---'



class LegacyCLI(Configurable):
    needs = {'storage': Storage}

    def find_facts(self, query, days=1, summary=False,
                   show_date_if_crosses_days=False, compact=False):
        return find_facts(self['storage'], query, days=1, summary=False,
                          show_date_if_crosses_days=False, compact=False)

    @arg('-c', '--continued', help='continue from last stop')
    def punch_in(self, activity, continued=False, interactive=False):
        return punch_in(self['storage'], activity, continued=False,
                        interactive=False)

    @named('out')
    @arg('-t', '--tags', help='comma-separated list of tags')
    @arg('-w', '--with_person', help='-w john,mary = -t with-john,with-mary')
    def punch_out(self, description=None, tags=None, with_person=None):
        return punch_out(self['storage'], description=None, tags=None,
                         with_person=None)

    def _complete_activity(self, prefix, **kwargs):
        # an argcomplete completer
        known = (x['activity'] for x in self.storage.get_known_activities())
        return [a for a in known if a.startswith(prefix)]

    @arg('activity', nargs='?', help='must be specified unless --amend is set',
     completer=_complete_activity)
    @arg('-a', '--amend', default=False,
        help='update last fact instead of creating a new one')
    @arg('-d', '--description')
    @arg('-t', '--tags', help='comma-separated list of tags')
    @arg('--date', help='date to which --since and --until are appended')
    @arg('-s', '--since', help='activity start time (HH:MM)')
    @arg('-u', '--until', help='activity end time (HH:MM)')
    @arg('--duration', help='activity duration (HH:MM)')
    @arg('-b', '--between', help='HH:MM-HH:MM')
    @arg('-w', '--with-person', help='-w john,mary = -t with-john,with-mary')
    @arg('--dry-run', default=False, help='do not alter the database')
    @arg('-p', '--pick', default=None, help='last activity name to pick if --amend flag '
        'is set (if not given, last activity is picked, regardless of its name)')
    @arg('--no-input', default=False, help='no additional interactive input')
    @named('log')
    @wrap_errors([StorageError, NotFoundError], processor=failure)
    @expects_obj
    def log(self, args):
        return log_activity(args)

    @property
    def commands(self):
        "Returns a list of functions to use as CLI commands"
        return [
            self.find_facts,
            self.punch_in,
            self.punch_out,
            self.log,
            # FIXME these are not configurable and therefore won't work:
            add_post_scriptum,
            show_last_fact,
            update_fact,
            load_from_file
        ]



#!/usr/bin/env python
# coding: utf-8
# PYTHON_ARGCOMPLETE_OK
#
#    Timetra is a time tracking application and library.
#    Copyright © 2010-2014  Andrey Mikhaylenko
#
#    This file is part of Timetra.
#
#    Timetra is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Lesser General Public License as published
#    by the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    Timetra is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public License
#    along with Timetra.  If not, see <http://gnu.org/licenses/>.
#
"""
~~~~~~~~~~~~~~~~~~~
Simple Diary Script
~~~~~~~~~~~~~~~~~~~

A simple temporary frontend for the YAML backend.

:author: Andrey Mikhaylenko

"""
import datetime
import subprocess

import argh
import blessings
from confu import Configurable

from .storage import Storage
from . import utils


t = blessings.Terminal()


FACT_FORMAT = ('{since.year}-{since.month:0>2}-{since.day:0>2} '
               '{since.hour:0>2}:{since.minute:0>2}-'
               '{until.hour:0>2}:{until.minute:0>2} '
               '{activity} {duration} {description}')


class Diary(Configurable):
    needs = {
        'storage': Storage,
    }

    def find(self, since=None, until=None, activity=None, note=None, tag=None,
             fmt=FACT_FORMAT):

        if since:
            since = datetime.datetime.strptime(since, '%Y-%m-%d')
        if until:
            until = datetime.datetime.strptime(until, '%Y-%m-%d')

        facts = self.storage.find(since=since, until=until, activity=activity,
                                  description=note, tag=tag)
        for fact in facts:
            fact['activity'] = t.yellow(fact['activity'])
            # avoid "None" in textual representation
            fact['description'] = t.blue(fact['description'] or '')
            try:
                delta = fact['until'] - fact['since']
                fact['duration'] = '{:.0f}m'.format(delta.total_seconds() / 60)
            except:
                fact['duration'] = ''

            yield fmt.format(**fact)


    def today(self):
        date = datetime.datetime.today().strftime('%Y-%m-%d')
        return self.find(since=date)


    def yesterday(self):
        date = datetime.datetime.today() - datetime.timedelta(days=1)
        return self.find(since=date.strftime('%Y-%m-%d'))


    def edit(self, date=None):
        if isinstance(date, (datetime.date, datetime.datetime)):
            pass
        elif date:
            date = datetime.datetime.strptime(date, '%Y-%m-%d')
        else:
            date = datetime.date.today()

        path = self.storage.backend.get_file_path_for_day(date)
        print('opening', path, 'in editor...')
        subprocess.Popen(['vim', path]).wait()
        print('editor finished.')


    @argh.wrap_errors([AssertionError])
    def add(self, activity, since, until=None, duration=None, note=None,
            tags=None, open_editor=False):
        if since == 'thru':
            prev = self['storage'].get_latest()
            since = prev['until']
        else:
            since = timetra.utils.parse_time_to_datetime(since)

        if until:
            assert not duration
            until = timetra.utils.parse_time_to_datetime(until)
        elif duration:
            until = since + datetime.timedelta(minutes=int(duration))
        else:
            until = datetime.datetime.now()

        assert since < until, '--since must be earlier than --until'

        fact = {
            'activity': activity,
            'since': since,
            'until': until,
            'description': note,
            'tags': tags.split(',') if tags else [],
        }
        file_path = self.storage.add(fact)

        print('Added {} +{:.0f}m to {}'.format(since.strftime('%Y-%m-%d %H:%M'),
                                    (until - since).total_seconds() / 60,
                                    file_path))

        if open_editor:
            self.edit(since)



# -*- coding: utf-8 -*-
#
#    Timetra is a time tracking application and library.
#    Copyright © 2010-2014  Andrey Mikhaylenko
#
#    This file is part of Timetra.
#
#    Timetra is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Lesser General Public License as published
#    by the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    Timetra is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public License
#    along with Timer.  If not, see <http://gnu.org/licenses/>.
#
"""
===============
Terminal colors
===============
"""
from functools import partial
import sys

from blessings import Terminal


PY3 = sys.version_info >= (3,)
if PY3:
    unicode = str


t = Terminal()


def colored(text, wrapper):
    return wrapper(unicode(text))


success = partial(colored, wrapper=t.green)
warning = partial(colored, wrapper=t.yellow)
failure = partial(colored, wrapper=t.red)



# -*- coding: utf-8 -*-
#
#    Timetra is a time tracking application and library.
#    Copyright © 2010-2012  Andrey Mikhaylenko
#
#    This file is part of Timetra.
#
#    Timetra is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Lesser General Public License as published
#    by the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    Timetra is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public License
#    along with Timer.  If not, see <http://gnu.org/licenses/>.
#
"""
Reporting
=========
"""
from confu import Configurable
import prettytable

from ..storage import Storage
from .. import formatdelta
from .drift import show_drift
from .prediction import predict_next_occurence


class Reporting(Configurable):
    needs = {
        'storage': Storage,
    }

    def drift(self, activity='sleep', days=7, shift=False):
        return show_drift(self['storage'], activity, days, shift)

    def predict(self, activity):
        """ Predicts next occurence of given activity.
        """
        guess = predict_next_occurence(self['storage'], activity)
        table = prettytable.PrettyTable()
        table.field_names = 'start', 'end', 'duration', 'ETA'
        table.add_row([
            guess['start'].strftime('%Y-%m-%d %H:%M'),
            guess['end'].strftime('%Y-%m-%d %H:%M'),
            formatdelta.render_delta(guess['duration']),
            '{0}{1}'.format('-' if guess['eta_is_negative'] else '+',
                            formatdelta.render_delta(guess['eta'])),
        ])
        return table



# coding: utf-8

# python
import datetime

# 3rd party
from monk import modeling
from monk.schema import optional


FACT = {
    'category': str,
    'activity': str,
    'since': datetime.datetime.now,
    'until': datetime.datetime,
    'tags': optional([str]),
    optional('hamster_fact_id'): int,    # legacy
    'description': optional(str),
}


class Model(modeling.TypedDictReprMixin,
            modeling.DotExpandedDictMixin,
            modeling.StructuredDictMixin,
            dict):

    def __init__(self, *args, **kwargs):
        super(Model, self).__init__(*args, **kwargs)
        self._insert_defaults()
        self._make_dot_expanded()


class Fact(Model):
    structure = FACT

    @property
    def duration(self):
        return (self.until or datetime.datetime.now()) - self.since



# -*- coding: utf-8 -*-
#
#    Timetra is a time tracking application and library.
#    Copyright © 2010-2014  Andrey Mikhaylenko
#
#    This file is part of Timetra.
#
#    Timetra is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Lesser General Public License as published
#    by the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    Timetra is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public License
#    along with Timer.  If not, see <http://gnu.org/licenses/>.
#
import os
import subprocess
from warnings import warn


# Visible notifications
try:
    import notify2 as notify_lib
except ImportError:
    try:
        import pynotify as notify_lib
    except ImportError:
        notify_lib = None
if notify_lib:
    notify_lib.init('timetra')
else:
    warn('Visible alerts are disabled')


# Audible notifications
#try:
#    sys.path.insert(0, '/home/andy/src')
##    import beeper   # custom script, see http://paste.pocoo.org/show/316/
##    import beeper_alsa
#    subprocess.Popen(['beep', '-l', '0'])
#except OSError:
#    warn('Simple audible alerts are disabled')
#    beep_enabled = False
#else:
#    beep_enabled = True


__all__ = ['beep', 'say', 'show']


def beep(*pairs):
    """Emits beeps using the "beep" program."""
    beeps = []
    for frequency, duration in pairs:
        beeps.extend(['-f', str(frequency), '-l', str(duration), '-n'])
    beeps.pop()   # remove the last "-n" separator to prevent extra beep
    subprocess.Popen(['beep'] + beeps) #, '-f', str(frequency), '-l', str(duration)])
    #'beep -f 100 -n -f 150 -n -f 50 -n -f 300 -n -f 200 -n -f 400'.split())
#    try:
#        beeper.beep(frequency, duration)
#    except IOError:
#        beeper_alsa.beep(frequency, duration)


def say(text):
    """Uses Festival TTS to actually say the message."""
    # see http://ubuntuforums.org/showthread.php?t=751169
    sound_wrapper = 'padsp'  # or 'aoss' or 'esddsp' or none
    command = 'echo \'(SayText "{text}")\' | {sound_wrapper} festival &'
    text = text.replace('"','').replace("'",'')
    os.system(command.format(sound_wrapper=sound_wrapper, text=text))


def show(text, critical=False):
    if not notify_lib:
        return False

    note = notify_lib.Notification(summary='Time Tracker', message=text)
    if critical:
        note.set_urgency(notify_lib.URGENCY_CRITICAL)
    note.show()
    return True



#!/usr/bin/env python2
# -*- coding: utf-8 -*-
#
#    Timetra is a time tracking application and library.
#    Copyright © 2010-2014  Andrey Mikhaylenko
#
#    This file is part of Timetra.
#
#    Timetra is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Lesser General Public License as published
#    by the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    Timetra is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public License
#    along with Timer.  If not, see <http://gnu.org/licenses/>.
#
import urwid


class TabNavigatedFrame(urwid.Frame):
    """ A :class:`urwid.Frame` that supports changing focus between its parts
    on :key:`tab`.
    """
    def keypress(self, size, key):
        switch_order = {
            'tab': {
                'header': 'body',
                'body': 'footer',
                'footer': 'header',
            },
            'shift tab': {
                'header': 'footer',
                'body': 'header',
                'footer': 'body',
            }
        }
        if key not in switch_order:
            return self.__super.keypress(size, key)
        next_part = switch_order[key][self.focus_part]
        self.set_focus(next_part)


class History(list):
    """ A list that remembers current position::

        >>> h = History()
        >>> h
        <History [] at 0>
        >>> h.up()
        >>> h.down()
        >>> h.append('a')
        >>> h
        <History ['a'] at 0>
        >>> h.up()
        'a'
        >>> h.down()
        'a'
        >>> h.append('b')
        >>> h
        <History ['a', 'b'] at 1>
        >>> h.up()   # first up
        'a'
        >>> h.up()   # second up
        'a'
        >>> h.down() # first down
        'b'
        >>> h.down() # second down
        'b'

    """
    def __init__(self, *args):
        super(History, self).__init__(*args)
        self.position = len(self) - 1 if self else 0

    def __repr__(self):
        return '<History {content} at {position}>'.format(
            content = super(History, self).__repr__(),
            position = self.position
        )

    def append(self, item):
        if not item:
            return
        if self and self[-1] == item:
            return
        super(History, self).append(item)
        # reset index to the last added value
        self.position = len(self) - 1

    def up(self):
        if not self:
            return
        if 0 < self.position:
            self.position -= 1
        return self[self.position]

    def down(self):
        if not self:
            return
        if self.position < len(self) - 1:
            self.position += 1
        return self[self.position]

    def get_current(self):
        if not self:
            return
        return self[self.position]


class Prompt(urwid.Edit):
    """ A :class:`urwid.Edit` that can be submitted with :key:`enter` to a
    special handler function. The field is cleared on submit.

    Usage (with Python3 syntax for shorter example)::

        def handle_command(raw_command):
            command, *args = ' '.split(raw_command)
            if command == 'help':
                show_help(args)
            elif cmd == 'save':
                save_as(args[0])

        prompt = Prompt(u'>>> ', controller=handle_command)

    """
    def __init__(self, *args, **kwargs):
        self.handle_value = kwargs.pop('controller')
        # TODO: ideally, the history should be stored in a file to recover last
        #       command in case of a crash.
        self.history = History()
        self.__super.__init__(*args, **kwargs)

    def keypress(self, size, key):
        if key == 'enter':
            self.history.append(self.edit_text)
            handled = self.handle_value(self.edit_text)
            if handled:
                self.edit_text = u''
            return
        elif key == 'up':
            if not self.edit_text:
                self.edit_text = self.history.get_current()
            else:
                self.edit_text = self.history.up()
        elif key == 'down':
            if not self.edit_text:
                self.edit_text = self.history.get_current()
            else:
                self.edit_text = self.history.down()

        return self.__super.keypress(size, key)



# -*- coding: utf-8 -*-
#
#    Timetra is a time tracking application and library.
#    Copyright © 2010-2014  Andrey Mikhaylenko
#
#    This file is part of Timetra.
#
#    Timetra is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Lesser General Public License as published
#    by the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    Timetra is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public License
#    along with Timer.  If not, see <http://gnu.org/licenses/>.
#
"""
Prediction
==========

:copyright: Andy Mikhaylenko, 2012
:license: LGPL3

.. note:: TODO

   READ THIS: http://otexts.com/fpp/

"""
from datetime import datetime, timedelta


def avg_delta(deltas):
    deltas_as_seconds = [delta.total_seconds() for delta in deltas]
    avg_seconds = sum(deltas_as_seconds) / float(len(deltas_as_seconds))
    return timedelta(seconds=avg_seconds)


def predict_next_occurence(storage, activity, num_facts=4):
    """ Returns a tuple `(start_time, end_time)` of the next expected occurence
    of given activity.

    The algo is pretty dumb: take last N facts, take average gap between the
    facts to guess next occurrence relative to the last known fact, and use
    average duration as estimated duration.
    """
    all_facts = storage.get_facts_for_day(date=-1, search_terms=activity)
    recent_facts = all_facts[-num_facts:]
    if len(recent_facts) < 2:
        return None
    gaps = []
    prev = None
    for f in recent_facts:
        if prev:
            gaps.append(f.start_time - prev.end_time)
        prev = f
    est_gap = avg_delta(gaps)
    est_start = recent_facts[-1].end_time + est_gap
    est_duration = avg_delta(f.delta for f in recent_facts)
    est_end = est_start + est_duration

    now = datetime.now()
    if now < est_start:
        eta = est_start - now
        eta_is_negative = False
    else:
        eta = now - est_start
        eta_is_negative = True
    return {'start': est_start, 'end': est_end, 'duration': est_duration,
            'eta': eta, 'eta_is_negative': eta_is_negative}



# -*- coding: utf-8 -*-
#
#    Timetra is a time tracking application and library.
#    Copyright © 2010-2014  Andrey Mikhaylenko
#
#    This file is part of Timetra.
#
#    Timetra is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Lesser General Public License as published
#    by the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    Timetra is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public License
#    along with Timer.  If not, see <http://gnu.org/licenses/>.
#
"""
Utility functions
=================
"""
from datetime import date, datetime, time, timedelta


def to_date(obj):
    if isinstance(obj, datetime):
        return obj.date()
    if isinstance(obj, date):
        return obj
    raise TypeError('expected date or datetime, got {0}'.format(obj))


def to_datetime(obj):
    if isinstance(obj, datetime):
        return obj
    if isinstance(obj, date):
        return datetime.combine(obj, time(0))
    raise TypeError('expected date or datetime, got {0}'.format(obj))



# TODO: use  https://bitbucket.org/russellballestrini/ago/src
#        or  https://github.com/tantalor/pretty_timedelta

def format_delta(delta, fmt='{hours}:{minutes}'):
    """ Formats timedelta. Allowed variable names are: `days`, `hours`,
    `minutes`, `seconds`.
    """
    hours, rem = divmod(delta.seconds, 3600)
    minutes, seconds = divmod(rem, 60)
    return fmt.format(days=delta.days, hours=hours, minutes=minutes,
                      seconds=seconds)


def split_time(string):
    """ Returns a pair of integers `(hours, minutes)` for given string::

        >>> _split_time('02:15')
        (2, 15)
        >>> _split_time('2:15')
        (2, 15)
        >>> _split_time(':15')
        (0, 15)
        >>> _split_time('15')
        (0, 15)

    """
    def _split(s):
        if ':' in s:
            return s.split(':')
        if len(s) <= 2:
            # "35" -> 00:35
            return 0, s
        return s[:-2], s[-2:]

    return tuple(int(x or 0) for x in _split(string))


def parse_time(string):
    """
    Returns a datetime.time object and a boolean that tells whether given time
    should be substracted from the start time.

        >>> parse_time('20')
        (time(0, 20), False)
        >>> parse_time('-1:35')
        (time(1, 35), True)
        >>> parse_time('now')
        datetime.now().time()
        >>> parse_time('now-5')
        datetime.now().time() - timedelta(minutes=5)

    """
    substract = False

    if string == 'now':
        return datetime.now().time(), substract

    if string.startswith('now-'):
        # "now-150"
        now = datetime.now()
        _, substring = string.split('-')
        delta, _ = parse_time(substring)
        start = now - timedelta(hours=delta.hour, minutes=delta.minute)
        return start.time(), substract

    if string.startswith('-'):
        # "-2:58"
        substract = True
        string = string[1:]

    hours, minutes = split_time(string)

    return time(hours, minutes), substract


def parse_time_to_datetime(string, relative_to=None, ensure_past_time=True):
    """ Parses string to a datetime, relative to given date (or current one):

    CURRENT FORMAT:
        12:05 = DATE, at 12:05
    TODO:
         1205 = DATE, at 12:05
          205 = DATE, at 02:05
           05 = DATE, at 00:05
            5 = DATE, at 00:05
           -5 = DATE - 5 minutes
    """
    if not string:
        return
    base_date = relative_to or datetime.now()
    parsed_time, _ = parse_time(string)
    date_time = datetime.combine(base_date, parsed_time)

    # microseconds are not important but may break the comparison below
    base_date = base_date.replace(microsecond=0)
    date_time = date_time.replace(microsecond=0)

    if ensure_past_time and base_date < date_time:
        return date_time - timedelta(days=1)
    else:
        return date_time


def parse_delta(string):
    """ Parses string to timedelta.
    """
    if not string:
        return
    hours, minutes = split_time(string)
    return timedelta(hours=hours, minutes=minutes)



# -*- coding: utf-8 -*-
#
#    Timetra is a time tracking application and library.
#    Copyright © 2010-2014  Andrey Mikhaylenko
#
#    This file is part of Timetra.
#
#    Timetra is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Lesser General Public License as published
#    by the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    Timetra is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public License
#    along with Timer.  If not, see <http://gnu.org/licenses/>.
#
import sys
import time
import datetime

import argh
from confu import Configurable

from . import notification
from .models import Fact
from .storage import Storage
from .term import success, warning, failure


PY3 = sys.version_info >= (3,)


# remind that "less than a minute left" each N seconds
LAST_MINUTE_ALARM_FREQUENCY = 30


class Period(object):
    ALARM_START = 1
    ALARM_END = 2
    ALARM_CANCEL = 3
    ALARM_REMIND = 4

    def __init__(self, minutes, name=None, category_name='workflow',
                 description='', trackable=False, tags=[], silent=False,
                 storage=None):
        self.minutes = int(minutes)
        self.name = name
        self.category_name = category_name
        self.description = description
        self.is_trackable = trackable
        self.tags = tags
        self.until = None
        self.silent = silent
        self.storage = storage
        self.current_fact = None

        if trackable and not storage:
            raise RuntimeError('storage if required with trackable=True')

    def __int__(self):
        return int(self.minutes)

    def __repr__(self):
        return str(self)

    def __str__(self):
        if self.name:
            return '{0.name} ({0.minutes} minutes)'.format(self)
        else:
            return '{0} minutes period'.format(int(self))

    def __unicode__(self):
        return str(self)

    def start(self):
        now = datetime.datetime.now()
        self.until = now + datetime.timedelta(minutes=self.minutes)

        if self.is_trackable:
            fact = Fact(activity=self.name, category=self.category_name,
                        description=self.description, tags=self.tags)
            self.storage.add(fact)
            self.current_fact = fact

#        message = 'Started {name} until {until}'.format(
        message = 'Started {name} for {minutes} minutes'.format(
            name = self.name or str(self),
#            until = self.until.strftime('%H:%M:%S'),
            minutes = self.minutes
        )
#        if until:
#            message += ' until {0}'.format(until.strftime('%H:%M:%S'))
        self.notify(message, self.ALARM_START)

    def stop(self):
        now = datetime.datetime.now()
        if self.until <= now:
            msg = '{self} is over!'.format(**locals())
            self.notify(msg, self.ALARM_END)
        else:
            delta = self.until - now
            msg = ('  Timer stopped with {0} minutes left (out of {1})'
                    ).format(int(delta.seconds / 60), self.minutes)
            self.notify(msg, self.ALARM_CANCEL)

        # TODO replace with signals
        if self.is_trackable:
            self.storage.update(self.current_fact,
                                {'until': datetime.datetime.now()})

        self.until = None

    def notify(self, message, mode=None, log=True, osd=True):
        if log:
            colored = failure if mode == self.ALARM_CANCEL else warning
            print('{time} {message}'.format(
                time = get_colored_now(),
                message = colored(message),
            ))

        if osd:
            notification.show(message, critical=bool(mode == self.ALARM_START))

        if not self.silent:
            beeps = []
            if mode == self.ALARM_START:
                beeps.append((350, 250))

                # emit an extra beep for every 5 minutes of the period
                # (this helps understand which period has just ended)
                for i in range(0, int(self.minutes / 5)):
                    # each beep is 50Hz higher than the previous one
                    freq = 50 * (i+1)
                    beeps.append((freq, 100))

                beeps.append((500,200))
            elif mode == self.ALARM_CANCEL:
                beeps += [(300, 50), (200, 50), (100, 200)]
            elif mode == self.ALARM_END:
                beeps += [(350, 250), (100, 200)]
            else:
                # probably the timer is about to end
                beeps += [(500, 20)]
            notification.beep(*beeps)

            # FIXME this should be used *instead* of beeping if Festival
            # program is available
            if mode in [self.ALARM_START, self.ALARM_REMIND,
                        self.ALARM_CANCEL]:
                notification.say(message)


def wait_for(period):
    until = datetime.datetime.now() + datetime.timedelta(minutes=int(period))
    period.start()
    while True:
        try:
            if until <= datetime.datetime.now():
                period.stop()
                return
            delta = until - datetime.datetime.now()
            if delta.seconds <= 60:
                period.notify('less than a minute left', period.ALARM_REMIND,
                              log=False)
            # wait a bit
            time.sleep(LAST_MINUTE_ALARM_FREQUENCY)
        except KeyboardInterrupt:
            period.stop()
            sys.exit()


def _once(*periods):
    for step in periods:
        wait_for(step)


def _cycle(*periods):
    print('Cycling periods: {0}'.format(', '.join([str(x) for x in periods])))
    while True:
        _once(*periods)


def get_colored_now():
    """Returns colored and formatted current time"""
    now = datetime.datetime.now().strftime('%Y-%m-%d %H:%M:%S')
    return success(now)


@argh.arg('periods', nargs='+')
def cycle(periods, silent=False):
    _cycle(*[Period(x, silent=silent) for x in periods])


@argh.arg('periods', nargs='+')
def once(periods, silent=False):
    _once(*[Period(x, silent=silent) for x in periods])


class Timing(Configurable):
    needs = {'storage': Storage}

    def pomodoro(self, activity='work', silent=False, work_duration=30,
                 rest_duration=10, description=''):
        yield 'Running Pomodoro timer'
        resolved = self['storage'].resolve_activity(activity)
        work_activity = resolved['activity']
        work_category = resolved['category']
        tags = ['pomodoro']

        work = Period(work_duration, name=work_activity,
                      category_name=work_category, trackable=True,
                      tags=tags, silent=silent,
                      description=description, storage=self['storage'])
        relax = Period(rest_duration, name='relax', trackable=True,
                       tags=tags, silent=silent, storage=self['storage'])

        _cycle(work, relax)




#
#    Timetra is a time tracking application and library.
#    Copyright © 2010-2014  Andrey Mikhaylenko
#
#    This file is part of Timetra.
#
#    Timetra is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Lesser General Public License as published
#    by the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    Timetra is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public License
#    along with Timer.  If not, see <http://gnu.org/licenses/>.
#
"""
Storage
=======
"""
#import datetime
import os
#from warnings import warn
import yaml


from . import caching, models


__all__ = ['Storage']


# http://stackoverflow.com/questions/8640959/how-can-i-control-what-scalar-form-pyyaml-uses-for-my-data
class literal(str):
    pass

def literal_representer(dumper, data):
    return dumper.represent_scalar('tag:yaml.org,2002:str', data, style='|')

yaml.add_representer(literal, literal_representer)


class YamlBackend:
    "Provides low-level access to the facts database"

    def __init__(self, data_dir, cache_dir):
        self.data_dir = data_dir
        self.cache = caching.Cache(cache_dir)

    def get_cached_day_file(self, path):
        return self.cache.get_cached_yaml_file(path, model=models.Fact)

    def _is_fact_matching(self, fact, filters):
        if not filters:
            return True
        for key, value in filters.items():
            if value not in (fact.get(key) or ''):
                return False
        return True

    def _collect_day_paths(self, since=None, until=None):

        for year in sorted(os.listdir(self.data_dir)):
            year_num = int(year)

            if since and year_num < since.year:
                continue
            if until and year_num > until.year:
                break

            year_path = os.path.join(self.data_dir, year)

            for month in sorted(os.listdir(year_path)):
                month_num = int(month)

                if since and year_num == since.year and month_num < since.month:
                    continue
                if until and year_num == until.year and month_num > until.month:
                    break

                month_path = os.path.join(year_path, month)

                for day in sorted(os.listdir(month_path)):
                    day_num = int(os.path.splitext(day)[0])

                    if since and year_num == since.year and month_num == since.month and day_num < since.day:
                        continue
                    if until and year_num == until.year and month_num == until.month and day_num > until.day:
                        break

                    yield os.path.join(month_path, day)

    def collect_facts(self, since=None, until=None, filters=None,
                      hint_reverse=False):
        day_paths = self._collect_day_paths(since=since, until=until)
        if hint_reverse:
            # optimization hint
            day_paths = reversed(list(day_paths))
        for day_path in day_paths:
            day_facts = self.get_cached_day_file(day_path)
            if hint_reverse:
                day_facts = reversed(day_facts)
            for fact in day_facts:
                if self._is_fact_matching(fact, filters):
                    yield fact

    def get_file_path_for_day(self, date):
        return os.path.join(
            self.data_dir,
            str(date.year),
            '{:0>2}'.format(date.month),
            '{:0>2}.yaml'.format(date.day),
        )

    def _load_from_file(self, file_path):
        if os.path.exists(file_path):
            with open(file_path) as f:
                return yaml.load(f)
        return []

    def _dump_to_file(self, file_path, facts, create=False):
        if not os.path.exists(file_path):
            # make sure the year and month dirs are created
            month_dir = os.path.dirname(file_path)
            if not os.path.exists(month_dir):
                os.makedirs(month_dir)

        with open(file_path, 'w') as f:
            yaml.dump(facts, f, allow_unicode=True, default_flow_style=False)

    def add(self, fact):
        # we expect the `fact` dictionary to be already validated
        file_path = self.get_file_path_for_day(fact['since'])
        facts = self._load_from_file(file_path)

        # convert to generic YAML mapping
        fact_data = dict(fact)

        inserted = False
        for i, other in enumerate(facts):
            if fact['since'] < other['since']:
                facts.insert(i, fact_data)
                inserted = True
                break
        if not inserted:
            facts.append(fact_data)

        # dump multi-line fields as literal blocks (the "|+" stuff)
        for f in facts:
            for field, value in f.items():
                if value and isinstance(value, str) and '\n' in value:
                    f[field] = literal(value)

        self._dump_to_file(file_path, facts, create=True)

        return file_path

    def get(self, date_time):
        facts = self.collect_facts(since=date_time.date(),
                                   until=date_time.date())
        for fact in facts:
            if fact['since'] == date_time:
                return fact

        raise FactNotFound(date_time)

    def delete(self, since, activity):
        file_path = self.get_file_path_for_day(since)
        facts = self._load_from_file(file_path)

        for i, fact in enumerate(facts):
            if fact['since'] == since and fact['activity'] == activity:
                facts.pop(i)
                break
        else:
            raise FactNotFound('{} {}'.format(since, activity))

        self._dump_to_file(file_path, facts, create=False)

    def update(self, old_fact, kwargs):
        # make sure it exists
        existing_fact = self.get(old_fact['since'])
        assert existing_fact['activity'] == old_fact['activity']

        new_fact = models.Fact(old_fact, **kwargs)
        new_fact.validate()

        # If date_time changes, we may need to delete from one file and
        # write to another one (i.e. 2 files are updated).
        #
        # To support this case without extra logic we simply delete old record
        # and then add a new one.
        #
        # The problem is that there's no transactions here; we cannot even
        # add before deleting because if the fact stays at the same place,
        # we'd have two similar facts one after another and it would be hard
        # to tell which one should be deleted.

        self.delete(old_fact['since'], old_fact['activity'])
        self.add(new_fact)

    def get_latest(self):
        return self.collect_facts(hint_reverse=True).__next__()

    def find(self, since=None, until=None, activity=None, description=None, tag=None):
        filters = {}
        if activity:
            filters['activity'] = activity
        if description:
            filters['description'] = description
        if tag:
            filters['tags'] = tag
        return self.collect_facts(since=since, until=until, filters=filters)


class Storage:
    "Provides high-level access to the facts database"

    def __init__(self, backend):
        self.backend = backend

    def get(self, date_time):
        return self.backend.get(date_time)

    def get_latest(self):
        return self.backend.get_latest()

    def find(self, since=None, until=None, activity=None, description=None,
             tag=None):
        return self.backend.find(since=since, until=until, activity=activity,
                                 description=description, tag=tag)

    def add(self, fact):
        """Adds given fact to the database.  Returns whatever the backend
        returned (presumably the fact ID or something that can find/identify
        the newly created record).
        """
        # TODO [not only] Monk-powered validation
        fields = ('activity', 'since', 'until', 'description', 'tags')
        assert all(x in fact for x in fields)

        # TODO: identify and solve conflicts
        #fact.validate()
        #if fact.since in self.backend:
        #    # TODO: check overlap
        #    raise FactsInConflict('another fact starts with this date and time')
        #self.backend[fact.since] = fact
        return self.backend.add(fact)

    def update(self, fact, values):
        assert fact
        assert values
        return self.backend.update(fact, values)

    def delete(self, spec):
        assert spec.since, spec.activity
        fact = self.get(spec.since)
        assert fact
        assert fact.activity == spec.activity
        self.backend.delete(since=fact.since, activity=fact.activity)

    def resolve_activity(self, mask):
        """Given a mask, finds the (single) matching activity and returns its full
        name along with category name. Raises AssertionError if no matching
        activity could be found or more than item matched.

        :return: {'category': CATEGORY, 'activity': ACTIVITY}
        """
        seen = {}
        for fact in self.find():
            pair = fact.activity, fact.category
            seen[pair] = seen.get(pair, 0) + 1

        # look for exact matches
        sorted_seen = [x for x in sorted(seen, key=seen.get, reverse=True)]
        candidates = [d for d in sorted_seen if mask == d[0]]
        if not candidates:
            # look for partial matches
            candidates = [d for d in sorted_seen if mask in d[0]]
        if not candidates:
            raise UnknownActivity('unknown activity {0}'.format(mask))

        if 1 < len(candidates):
            # okay, too many; let's exclude empty categories
            candidates_with_categories = [c for c in candidates if c[1]]
            if candidates_with_categories:
                candidates = candidates_with_categories

        if 1 < len(candidates):
            raise AmbiguousActivityName('ambiguous name, matches: {0}'.format(
                '; '.join(('{1}/{0}'.format(*x)
                           for x in sorted(candidates)))))
        activity, category = candidates[0]
        return {'activity': activity, 'category': category}

    def get_known_activities(self):
        return self.backend.get_known_activities()


#--------------
# Auxiliary API
#

class StorageError(Exception):
    """ Base class for all storage-related exceptions.
    """


class ActivityMatchingError(StorageError):
    """ Raised if no known activity unambiguously matches given pattern.
    """


class UnknownActivity(ActivityMatchingError):
    """ Raised if no activity in the storage corresponds to given pattern.
    """


class AmbiguousActivityName(ActivityMatchingError):
    """ Raised if more than a single known activity matches given pattern.
    """


class CannotCreateFact(StorageError):
    pass


class FactNotFound(StorageError):
    pass


class FactsInConflict(StorageError):
    pass


'''
def get_start_end(since, until, delta):
    """
    :param since: `datetime.datetime`
    :param until: `datetime.datetime`
    :param delta: `datetime.timedelta`

    * since .. until
    * since .. since+delta
    * since .. now
    * until-delta .. until
    * prev .. until
    * prev .. prev+delta
    * prev .. now

    """
    prev = get_prev_end_time(require=True)
    now = datetime.datetime.now()

    if since and until:
        return since, until
    elif since:
        if delta:
            return since, since+delta
        else:
            return since, now
    elif until:
        if delta:
            return until-delta, until
        else:
            return prev, until
    elif delta:
        return prev, prev+delta
    else:
        return prev, now
'''



# coding: utf-8
import logging
import os
import shelve

import yaml
from monk.errors import ValidationError
from monk.validation import validate


__all__ = ['Cache']


log = logging.getLogger(__name__)


class Cache:
    APP_NAME = 'timetra'
    FILE_NAME = 'yaml_files.db'

    def __init__(self, root_dir=None):
        cache_dir = root_dir or self._make_xdg_dir()
        path = cache_dir + '/' + self.FILE_NAME

        try:
            db = shelve.open(path)
        except:
            os.remove(path)
            db = shelve.open(path)

        self.path = path
        self.db = db

    def _make_xdg_dir(self):
        import xdg.BaseDirectory
        return xdg.BaseDirectory.save_cache_path(self.APP_NAME)

    def get_cached_yaml_file(self, path, model):
        #results = tmpl_cache.get(key=search_param, createfunc=load_card)
        time_key = 'changed:' + path
        data_key = 'content:' + path
        mtime_cache = self.db.get(time_key)
        mtime_file = os.stat(path).st_mtime
        if mtime_cache == mtime_file:
            data = self.db[data_key]
            log.debug('[x]', path)
        else:
            log.debug('[ ]', path)
            data = list(self._load_object_list(path, model))
            self.db[data_key] = data
            self.db[time_key] = mtime_file
        #cache.close()
        return data


    def _load_object_list(self, path, model):
        with open(path) as f:
            try:
                items = yaml.load(f)
            except:
                print('FAILED to load', model, 'from', path)
                raise

        if not items:
            return

        for data in items:
            obj = model(data)

            try:
                validate(model, obj)
            except (ValidationError, TypeError) as e:
                raise type(e)('{path}: {e}'.format(path=path, e=e))

            yield obj

    def reset(self):
        try:
            self.db.close()
        except:
            pass
        os.remove(self.path)


#cache = Cache()
#
#commands = [ cache.reset ]



#!/usr/bin/env python
# coding: utf-8
# PYTHON_ARGCOMPLETE_OK
#
#    Timetra is a time tracking application and library.
#    Copyright © 2010-2014  Andrey Mikhaylenko
#
#    This file is part of Timetra.
#
#    Timetra is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Lesser General Public License as published
#    by the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    Timetra is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public License
#    along with Timetra.  If not, see <http://gnu.org/licenses/>.
#
"""
~~~~~~~~~~~~~~~~~~~
Timetra Application
~~~~~~~~~~~~~~~~~~~

:author: Andrey Mikhaylenko

"""
import os

import argh
import yaml

from .storage import Storage, YamlBackend

from .diary import Diary
from .reporting import Reporting
from .timer import Timing
from .curses import TUI
#from .cli_old import LegacyCLI


CONF_FILE = os.getenv('TIMETRA_CONFIG', 'conf.yaml')


def _init_storage():
    with open(CONF_FILE) as f:
        conf = yaml.load(f)
    backend = YamlBackend(**conf['backend'])
    storage = Storage(backend)
    return storage


def main():
    storage = _init_storage()

    p = argh.ArghParser()

    diary = Diary({'storage': storage})
    reporting = Reporting({'storage': storage})
    timing = Timing({'storage': storage})
    tui = TUI({'storage': storage})
    #old_cli = LegacyCLI({'storage': storage})

    command_tree = {
        None: [
            diary.find, diary.add, diary.edit,
            diary.today, diary.yesterday
        ],
        'report': [
            reporting.drift
        ],
        'timing': [
            timing.pomodoro
        ],
        'tui': [
            tui.run
        ],
        #'old':    old_cli.commands,
    }
    for namespace, commands in command_tree.items():
        p.add_commands(commands, namespace=namespace)

    p.dispatch()


if __name__ == '__main__':
    main()



#!/usr/bin/env python2
# -*- coding: utf-8 -*-
#
#    Timetra is a time tracking application and library.
#    Copyright © 2010-2014  Andrey Mikhaylenko
#
#    This file is part of Timetra.
#
#    Timetra is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Lesser General Public License as published
#    by the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    Timetra is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public License
#    along with Timer.  If not, see <http://gnu.org/licenses/>.
#
"""
Curses UI
=========

:author: Andrey Mikhaylenko
:dependencies:
    * urwid_

.. _urwid: http://excess.org/urwid/
"""
import datetime
from functools import partial
import shlex
import sys

from confu import Configurable
import urwid

#from timetra import timer
from ..storage import Storage
from . import widgets


PY3 = sys.version_info >= (3,)
if PY3:
    unicode = str


# TODO: move to config
#
# TODO: procrastination is not a property of an activity on its own;
#       it's rather a property of an activity in the context of current
#       priorities and the possibilities to tackle higher-priority tasks.
#       That is, the user is procrastinating if current activity has lower
#       priority than a number of planned actions doable right now.
#
#       In short:
#
#       * a "good" activity is bound to a high-priority task (HPT);
#       * a "bad" activity is anything that prevents doing a HPT that is
#         relevant in given context;
#       * if there's no relevant HPT at the moment, any activity is OK.
#
CATEGORY_COLOURS = {
    # "desirable" categories
    'productive': ('education', 'errands', 'self-care', 'work', 'maintenance'),
    # "undesirable" categories
    'procrastination': ('needless'),
}

def format_time(date_time):
    if not date_time:
        return u'.....'
    return date_time.strftime('%H:%M')


def get_delta_graph(delta_seconds, minutes_in_block=10):
    blocks_cnt = int(delta_seconds / 60 / minutes_in_block) or 1
    return '■' * blocks_cnt


def get_colour(category):
    if not category:
        return
    for colour in CATEGORY_COLOURS:
        if category in CATEGORY_COLOURS[colour]:
            return colour


class DayView(object):
    palette = [
        ('body',            'black',       'light gray', 'standout'),
        ('input normal',    'white',       'black'),
        ('input select',    'black',       'yellow'),
        ('button',          'light gray',  'dark green', 'standout'),
        ('button select',   'white',       'dark green'),
        ('header',          'white',       'dark blue'),
        ('bg background',   'light gray',  'black'),
        ('bg 1',            'black',       'dark blue', 'standout'),
        ('bg 2',            'black',       'dark cyan', 'standout'),
        ('productive',      'light green', 'default'),
        ('procrastination', 'light red',   'default'),
        ('cmd_output',      'light blue',  'default'),
        ('prompt',          'light gray',  'default'),
    ]

    def __init__(self, storage):
        self.storage = storage

        self.factlog = urwid.ListBox(urwid.SimpleListWalker([]))
        self.stats = urwid.ListBox(urwid.SimpleListWalker([]))

        prompt = widgets.Prompt(u':', controller=self.handle_prompt)
        self.prompt = urwid.AttrWrap(prompt, 'prompt')

        header = urwid.Text(u'(not refreshed)')
        body = urwid.Columns([
            self.factlog,
            self.stats,
        ])

        cmd_output = urwid.Text(u'')    # command output
        cmd_output = urwid.AttrWrap(cmd_output, 'cmd_output')

        footer = urwid.Pile([
            cmd_output,
            self.prompt
        ])
        self.frame = widgets.TabNavigatedFrame(header=header, body=body, footer=footer)
        self.refresh_data()

    def run(self):
        loop = urwid.MainLoop(self.render(), self.palette,
                              unhandled_input=self.unhandled_input)
        self.set_refresh_timeout(loop)
        loop.run()

    def set_refresh_timeout(self, main_loop, user_data=None):
        self.refresh_data()
        main_loop.set_alarm_in(2, self.set_refresh_timeout)

    def refresh_factlist(self, facts):
        # Activity log

        # replace list items without dropping the list object
        # (so self.factlog.body doesn't get orphaned)
        self.factlog.body[:] = []

        for fact in facts:
            delta_graph = unicode(get_delta_graph(fact.duration.total_seconds()))
            text = urwid.Text([
                format_time(fact.since),
                u'  {0: >2.0f}  '.format(fact.duration.total_seconds() / 60),
                (get_colour(fact.category), delta_graph),
                u'  ',
                fact.activity,
            ])
            self.factlog.body.append(text)

    def refresh_stats(self, facts):
        # Summary by categories
        self.stats.body[:] = []
        categories = {}
        if not facts:
            return
        for fact in facts:
            categories.setdefault(fact.category, datetime.timedelta(0))
            categories[fact.category] += fact.duration
        padding = max(len(str(x)) for x in categories if str(x)) + 1
        for category in sorted(categories, key=lambda k: categories[k]):
            total_seconds = categories[category].total_seconds()
            #tmpl = r'{category} {delta_graph}'.format()
            text = urwid.Text('{category: >{padding}} {delta_graph}'.format(
                    category = category,
                    delta_graph = get_delta_graph(total_seconds),
                    padding = padding,
            ))
            colour = get_colour(category)
            text = urwid.AttrWrap(text, colour)
            self.stats.body.append(text)

    def refresh_current_activity(self, facts):
        # Current activity
        now_text, now_delta = 'no activity', ''
        if facts:
            fact = facts[0]
            if fact.until:
                # display untracked time after last logged fact
                end_delta = datetime.datetime.now() - fact.until
                now_delta = u'({end_delta} minutes not tracked) '.format(
                    end_delta = end_delta.seconds / 60,
                )
            else:
                # display current activity
                now_text = u' {f.category}: {f.activity}'.format(f=fact)
                now_delta = u'{delta_graph} {f.delta} '.format(
                    f = fact,
                    delta_graph = get_delta_graph(fact.duration.total_seconds()),
                )

        head = urwid.Columns([
            urwid.Text(now_text),
            urwid.Text(now_delta, align='right')
        ])
        head = urwid.AttrWrap(head, 'header')
        self.frame.header = head

    def refresh_data(self):

        today = datetime.datetime.today()
        facts = list(reversed(list(self.storage.find(since=today))))

        if not facts:
            facts = [self.storage.get_latest_fact()]

        self.refresh_factlist(facts)
        self.refresh_stats(facts)
        self.refresh_current_activity(facts)

    def render(self):
        return self.frame

    def resume_activity(self):
        facts = self.storage.get_facts_for_day()   # XXX what if not today?
        if not facts:
            return
        fact = facts[-1]
        fact.until = None
        self.storage.update_fact(fact.id, fact)#, extra_description=comment)
        self.refresh_data()

    def quit(self):
        raise urwid.ExitMainLoop()

    def unhandled_input(self, key):
        if key == 'q':
            self.quit()
        elif key == ':':
            self.frame.set_focus('footer')

    def handle_timer_api(self, *argv):
        raise NotImplementedError
#        parser = app.ArghParser()
#        parser.add_commands(timer.commands)
#
#        response = u''
#        try:
#            response = parser.dispatch(argv, output_file=None)
#        except SystemExit:
#            pass
#        self.show_command_output(response)  # + unicode(args))

    def show_command_output(self, value):
        self.frame.footer.widget_list[0].set_text(unicode(value))

    def handle_prompt(self, value):
        # shlex.split respects quotes.
        # It does not however support Unicode prior to Python 2.7.3.
        if PY3:
            argv = shlex.split(value)
        else:
            argv = shlex.split(value.encode('utf-8'))
            argv = [unicode(arg) for arg in argv]
        if not argv:
            self.show_command_output(u'')
            return
        command = argv[0]

        def show_help(mapping, shortcuts):
            names = sorted(mapping)
            items = []
            for name in names:
                if name in shortcuts:
                    item = '{0} ({1})'.format(name, ' '.join(shortcuts[name]))
                else:
                    item = name
                items.append(item)
            self.show_command_output('  '.join(items))

        mapping = {
            'start':  partial(self.handle_timer_api, 'in'),
            'stop':   partial(self.handle_timer_api, 'out'),
            'resume': self.resume_activity,
            'log':    partial(self.handle_timer_api, 'log'),
            'last':   partial(self.handle_timer_api, 'show-last'),
            'quit':   self.quit,
            'timer':  self.handle_timer_api,
            'help':   'PLACEHOLDER',
        }
        shortcuts = {
            'start':  ['o'],
            'stop':   ['x'],
            'resume': ['r'],
            'log':    ['l'],
            'quit':   ['q', 'exit'],
            'timer':  ['t'],
            'help':   ['?'],
        }
        mapping['help'] = partial(show_help, list(mapping), shortcuts)
        for cmd, aliases in shortcuts.items():
            for alias in aliases:
                mapping[alias] = mapping[cmd]
        if command not in mapping:
            self.show_command_output(u'Unknown command "{0}"'.format(command))
            return
        func = mapping[command]
        func(*argv[1:])
        return True


class TUI(Configurable):
    "Timetra TUI (curses/urwid)"
    needs = {'storage': Storage}

    def run(self):
        view = DayView(storage=self['storage'])
        view.run()



# -*- coding: utf-8 -*-
#
#    Timetra is a time tracking application and library.
#    Copyright © 2010-2014  Andrey Mikhaylenko
#
#    This file is part of Timetra.
#
#    Timetra is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Lesser General Public License as published
#    by the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    Timetra is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public License
#    along with Timer.  If not, see <http://gnu.org/licenses/>.
#
"""
Methodology
===========

:copyright: Andy Mikhaylenko, 2012
:license: LGPL3

"""
import datetime


def check_planning_after_sleep(facts):
    remainder = []
    after_sleep = None
    sleep_fact = None
    for f in reversed(facts):
        if f.activity == 'sleeping':
            sleep_fact = f
            after_sleep = reversed(remainder)
            break
        remainder.append(f)
    message = u'Please plan the day'
    category = 'warning'
    if sleep_fact:
        if 60 * 30 < (datetime.datetime.now() - sleep_fact.end_time).total_seconds():
            category = 'error'
    if after_sleep:
        if any(True for x in after_sleep if x.activity == 'sorting-tasks'):
            message = u'You seem to have planned the day, OK'
            category = 'success'
    return message, category


def analyse_day(storage, date=None):
    """ Returns a generator or messages that give some hints re given day.
    """
    facts = storage.get_facts_for_day(date)
    rules = [check_planning_after_sleep]
    for rule in rules:
        yield rule(facts)



#!/usr/bin/env python
# coding: utf-8
#
#    Timetra is a time tracking application and library.
#    Copyright © 2010-2014  Andrey Mikhaylenko
#
#    This file is part of Timetra.
#
#    Timetra is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Lesser General Public License as published
#    by the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    Timetra is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public License
#    along with Timetra.  If not, see <http://gnu.org/licenses/>.
#

import io
import os

from setuptools import setup, find_packages

with io.open(os.path.join(os.path.dirname(__file__), 'README'), encoding='ascii') as f:
    readme = f.read()


try:
    from setuptools.command.easy_install import ScriptWriter
except ImportError:
    pass
else:
    # monkey-patch script writer template to enable bash completion
    # in 'console_scripts' entries
    ScriptWriter.template = '# PYTHON_ARGCOMPLETE_OK\n' + ScriptWriter.template


from timetra.diary import __version__


setup(
    # overview
    name             = 'timetra.diary',
    description      = 'Diary with CLI + YAML',
    long_description = readme,

    # technical info
    version  = __version__,
    packages = find_packages(),
    #provides = ['diary'],
    install_requires = [
        'argh>=0.22',
        'confu>=0.0.1',
        'monk>=0.11.2',
        'prettytable>=0.6.1',
        'python-dateutil>=2.1',
        'pyyaml>=3.10'
    ],
    # usage example: http://stackoverflow.com/a/18879288/68097
    extras_require = {
        # curses TUI
        'curses': [
            'urwid==1.1.1',
        ],
    },
    include_package_data = True,
    zip_safe = False,
        entry_points = {
        'console_scripts': [
            'timetra-diary=timetra.diary.app:main'
        ],
    },

    # copyright
    author   = 'Andrey Mikhaylenko',
    author_email = 'neithere@gmail.com',
    license  = 'GNU Lesser General Public License (LGPL), Version 3',

    # more info
    url          = 'https://github.com/neithere/timetra.diary',
    download_url = 'https://github.com/neithere/timetra.diary/archive/master.zip',

    # categorization
    keywords     = ('cli command line time tracking timer diary'),
    classifiers  = [
        'Development Status :: 3 - Alpha',
        'Environment :: Console',
        'Intended Audience :: Developers',
        'Intended Audience :: End Users/Desktop',
        'License :: OSI Approved :: GNU Library or Lesser General Public License (LGPL)',
        'Programming Language :: Python',
        'Topic :: Utilities',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ],
)



# -*- coding: utf-8 -*-
from datetime import datetime
from dateutil.relativedelta import relativedelta
from collections import namedtuple

from .utils import to_datetime


Part = namedtuple('Part', 'value name')


#finite_parts = dict(
#    months = 12,
#    days = 30,
#    hours = 24,
#    minutes = 60,
#)

naive_i18n = dict(
    years = u'лет',
    months = u'месяцев',
    days = u'дней',
    hours = u'часов',
    minutes = u'минут'
)


def format_delta_part(part):
    label = naive_i18n[part.name]
    return u'{value} {label}'.format(label=label, value=part.value)


def timedelta_to_relativedelta(td):
    minutes = hours = days = months = years = 0

    rem = int(td.total_seconds())

    seconds = rem % 60
    rem = rem // 60

    minutes = rem % 60
    rem = rem // 60

    hours = rem % 60
    rem = rem // 60

    days = rem % 24
    rem = rem // 24

    months = rem % 30
    rem = rem // 30

    years = rem % 12

    return relativedelta(years=years, months=months, days=days, hours=hours,
                         minutes=minutes, seconds=seconds)


def render_delta(dates_or_delta, stack_depth=2):
    """
    Renders pretty delta.  Usage::

        render_delta(delta)
        render_delta((d1, d2))

    """
    if isinstance(dates_or_delta, (tuple, list)):
        dt1, dt2 = dates_or_delta
        dt1 = to_datetime(dt1)
        dt2 = to_datetime(dt2)
        delta = relativedelta(dt2, dt1)
    else:
        delta = timedelta_to_relativedelta(dates_or_delta)

    stack = []

    if delta.years:
        stack.append(Part(delta.years, 'years'))

    if delta.months:
        stack.append(Part(delta.months, 'months'))

    if delta.days:
        stack.append(Part(delta.days, 'days'))

    if delta.hours:
        stack.append(Part(delta.hours, 'hours'))

    if delta.minutes:
        stack.append(Part(delta.minutes, 'minutes'))

    stack_cut = stack[:stack_depth]

#    def _filter_parts():
#        for part in reversed(stack_cut):
#            if part.name in finite_parts:
#                threshold = finite_parts[part.name] / 3
#                if threshold < part.value:
#                    yield part
#            else:
#                yield part
#    parts = list(reversed(list(_filter_parts())))

    parts = stack_cut
    if parts:
        return ' '.join(format_delta_part(p) for p in parts)
    else:
        return u'меньше минуты'
        #return u'только что'


if __name__ == '__main__':
    dt = datetime
    print(render_delta(dt(2012,6,20,19,0,0), dt(2012,6,20,19,1,0)))
    print(render_delta(dt(2012,6,20,14,0,0), dt(2012,6,20,19,1,0)))
    print(render_delta(dt(2012,6,19,18,0,0), dt(2012,6,20,19,1,0)))
    print(render_delta(dt(2012,5,19,18,0,0), dt(2012,6,20,19,1,0)))
    print(render_delta(dt(2012,3,19,18,0,0), dt(2012,6,20,19,1,0)))
    print(render_delta(dt(2012,1,19,18,0,0), dt(2012,6,20,19,1,0)))
    print(render_delta(dt(2011,3,19,18,0,0), dt(2012,6,20,19,1,0)))



#!/usr/bin/env python2
# -*- coding: utf-8 -*-
#
#    Timetra is a time tracking application and library.
#    Copyright © 2010-2014  Andrey Mikhaylenko
#
#    This file is part of Timetra.
#
#    Timetra is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Lesser General Public License as published
#    by the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    Timetra is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public License
#    along with Timer.  If not, see <http://gnu.org/licenses/>.
#
"""
Displays daily activity drift.

:copyright: Andy Mikhaylenko, 2012
:license: LGPL3
"""
import math
from datetime import datetime, timedelta

from prettytable import PrettyTable

from .. import utils


MARKER_EMPTY = '‧'
MARKER_FACTS = '■'
MARKER_NOW = '▹'  #'◉'  #'◗'

MIN_HOURLY_DURATION = 10
""" Minimum duration (in minutes) per hour. If the total duration of an
activity within given hour-long period is lower than this threshold, such
period is considered empty in regard to given activity.

.. note::

   "Within an hour-long period" != "within an hour".
   If a fact is split between two periods, it may disappear from results even
   if its total length exceeds the threshold.

"""


class HourData(object):
    def __init__(self, date, hour):
        self.date = date
        self.hour = hour
        self.duration = timedelta()

    def __repr__(self):
        return ('<{0.__class__.__name__} {0.date} '
                '{0.hour}h ({0.duration})>').format(self)

    def __str__(self):
        if self.is_current:
            return MARKER_NOW
        if timedelta(minutes=MIN_HOURLY_DURATION) < self.duration:
            return MARKER_FACTS
        return MARKER_EMPTY

    def __unicode__(self):
        return str(self)

    @property
    def is_current(self):
        now = datetime.now()
        if self.date == now.date() and self.hour == now.hour:
            return True
        else:
            return False


class DayData(list):
    "A list of hours (0..23) within a certain date."
    def __init__(self, date):
        self.date = date
        self[:] = [HourData(date, x) for x in range(24)]
        self.fact_cnt = 0
        self.min_start = None
        self.max_end = None

    @property
    def duration(self):
        hourly_durations = (x.duration for x in self)
        return sum(hourly_durations, timedelta())


class DriftData(dict):
    def __init__(self, span_days, end_time):
        for i in range(span_days):
            date = (end_time - timedelta(days=i)).date()
            self.ensure_date(date)

    def ensure_date(self, date):
        if date in self:
            return
        self[date] = DayData(date)

    def add_fact(self, start_time, end_time):
        pos = start_time.replace(minute=0, second=0, microsecond=0)
        while pos <= end_time:
            date = pos.date()
            self.ensure_date(date)
            if pos < start_time:
                # the fact starts somewhere between the edges
                right_edge = pos + timedelta(hours=1)
                if end_time <= right_edge:
                    # the fact completely fits this box
                    duration = end_time - start_time
                else:
                    # the fact overlaps the right edge
                    duration = right_edge - start_time
            else:
                # fact starts exactly at the hour edge
                duration = timedelta(hours=1)
                if end_time < pos + duration:
                    duration = end_time - pos
            self[date][pos.hour].duration += duration
            pos += timedelta(hours=1)

        day = self[date]
        day.fact_cnt += 1
        if not day.min_start or start_time < day.min_start:
            day.min_start = start_time
        if not day.max_end or day.max_end < end_time:
            if start_time.date() == end_time.date():
                day.max_end = end_time
            else:
                day.max_end = start_time.replace(
                    hour=23, minute=59, second=59, microsecond=0)


def collect_drift_data(storage, activity, span_days):
    span_days = span_days - 1  # otherwise it's zero-based
    until = datetime.now()
    since = until - timedelta(days=span_days)

    dates = DriftData(span_days, until)

    facts = storage.find(since, until=until, activity=activity)
    for fact in facts:
        dates.add_fact(fact.since, fact.until)

    return dates


def show_drift(storage, activity, days=7, shift=False):
    """Displays hourly chart for given activity for a number of days.
    Primary use: evaluate regularity of certain activity, detect deviations,
    trends, cycles. Initial intention was to find out my sleeping drift.
    """
    dates = collect_drift_data(storage, activity=activity, span_days=days)

    table = PrettyTable()

    fields = ['date', 'graph', 'total', 'total:graph']
    if shift:
        # diffs against previous day
        fields += ['qt', 'start', 'end']

    # prettytable would break if we appended items directly to this attr
    table.field_names = fields

    #table.align['start'] = table.align['end'] = 'r'
    table.align['total:graph'] = 'l'

    prev_day = None

    for date in sorted(dates):
        marks = dates[date]
        day = dates[date]
        spent = utils.format_delta(day.duration,
                                   fmt='{hours}:{minutes:0>2}')
        spent_graph = MARKER_FACTS * int(round(day.duration.total_seconds() / 60 / 60))

        if shift:
            shift_cnt = None
            shift_cnt_msg = shift_start_msg = shift_end_msg = ''
            if prev_day:
                shift_cnt = day.fact_cnt - prev_day.fact_cnt or ''
                if shift_cnt:
                    char = '+' if 0 < shift_cnt else '-'
                    shift_cnt_msg = char * int(math.copysign(shift_cnt, 1))

                shift_start_msg = get_shift_msg(day.min_start, prev_day.min_start)
                shift_end_msg = get_shift_msg(day.max_end, prev_day.max_end)

            shift_cells = [
                shift_cnt_msg,
                shift_start_msg or '',
                shift_end_msg or '',
            ]
        else:
            shift_cells = []

        table.add_row([
            date,
            ''.join((str(m) for m in marks)),
            spent,
            spent_graph,
        ] + shift_cells)

        prev_day = day

    return table


def get_shift_msg(dt1, dt2):
    if not dt1 or not dt2:
        return

    # we only need to compare time, not days
    dt2 = datetime.combine(dt1.date(), dt2.time())

    if dt1 < dt2:
        dt1, dt2 = dt2, dt1
        char = '◂'
    else:
        char = '▸'

    delta = dt1 - dt2

    if not delta:
        return

    delta_formatted = char * int(round(delta.total_seconds() // 60 / 60))
    return delta_formatted
